 /*---------------------------------------------------------------------------
                  Copyright (c) 2005 - 2007 demmel products
         This example demo code is provided as is and has no warranty,
     implied or otherwise.  You are free to use/modify any of the provided
    code at your own risk in your applications with the expressed limitation
    of liability (see below) so long as your product using the code contains
                      at least one uPSD products (device).

    LIMITATION OF LIABILITY:   NEITHER demmel products NOR ITS VENDORS OR
   AGENTS SHALL BE LIABLE FOR ANY LOSS OF PROFITS, LOSS OF USE, LOSS OF DATA,
       INTERRUPTION OF BUSINESS, NOR FOR INDIRECT, SPECIAL, INCIDENTAL OR
       CONSEQUENTIAL DAMAGES OF ANY KIND WHETHER UNDER THIS AGREEMENT OR
         OTHERWISE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
  --------------------------------------------------------------------------*/

 /*---------------------------------------------------------------------------

	Version 1.1

   This file should give you a good starting point for working with the iLCD's
	I2C port. All possible I2C commands (according to the specification in
	the "iLCD Commands" documentation, chapter "Controlling the iLCD Controller
   via I2C" are implemented in this file.

   It compiles on Keil-C for 8051, but it can be easily transferred to a
   different compiler and/or micro-controller.

	IMPORTANT NOTE: Ensure your iLCD controller fully supports I2C
	DPC10xx greater version 2.04, DPC20xx and DPC30xx greater or equal
   version 1.13

	PLEASE ADAPT THE VALUES IN HARDWARE.H TO YOUR NEEDS !!!

  --------------------------------------------------------------------------*/

#include <intrins.h>
#include "hardware.h"
#include "i2c.h"

// The iLCD controller can be driven by a 400kHz I2C signal, so the minumum clock high time
// is 1.25 us. The low level routines below use consecutive 1 cycle commands for setting/clearing
// the clock pin, so when this code is running on a standard 8051 derivate which devides the oscillator
// frequency by 12 a maximum frequency of 20 Mhz can be used with a delay of 2 cycles (realized via a nop)
// between this two consecutive set/clear port pin instructions. If the 8051 controller in use does not
// devide the clock by 12, adapt the delay() macro below accordingly.

#if FREQ_OSC > 20000
// enough for oscillator frequencies of up to 40 MHz
#define delay()	_nop_(); _nop_(); _nop_()		
#else
#define delay()	_nop(); _nop_()
#endif

#define I2C_RW					0x01		// OR'ed with i2c_address for reading

// Possible command codes for talking to the iLCD controller
#define CC_STATUS_READ		0x01		// get status byte from iLCD - bit definements see I2C.H
#define CC_DATA_SIZE_READ  0x02		// get # of bytes available for reading
#define CC_GET_INFO			0x03		// A combined command for getting status and # bytes
												// available for reading, returned in two consecutive bytes
#define CC_BLOCK_READ		0x04		// read a block of data from the iLCD controller
#define CC_BLOCK_WRITE		0x05		// write a block of data to the iLCD controller

// forward declarations of static functions
static byte i2c_receive_byte();
static word i2c_receive_word();
static byte i2c_shift_in();
static bool i2c_shift_out(byte data_byte);
static void i2c_ack();
static void i2c_nak();
static void i2c_stop();
static bool i2c_start();
static byte i2c_receive_byte();
static word i2c_receive_word();

bool i2c_result;
static byte i2c_address;

//----------------------------------------------------------------------------
// Initialize I2C
//----------------------------------------------------------------------------
void i2c_init(byte i2c_addr)
{
	i2c_address = i2c_addr;

	SDA_PIN = 1;
	SCL_PIN = 1;
}
//----------------------------------------------------------------------------
// Read status from the iLCD controller. i2c_result set to false in case of error.
// Done via a read byte protocol due to the I2C bus specification.
//----------------------------------------------------------------------------
byte i2c_read_status()
{
	// Output start condition
	if (!i2c_start())
		return 0;

	// Output I2C address (R/W bit low)
	if (!i2c_shift_out(i2c_address))
		return 0;

	// Output command CC_STATUS_READ
	if (!i2c_shift_out(CC_STATUS_READ))
		return 0;

	// As we have setup the read status mode now, get the status itself
	return i2c_receive_byte();
}
//----------------------------------------------------------------------------
// Read # of bytes contrained in output queue of the iLCD controller.
// i2c_result set to false in case of error.
// Done via a read byte protocol due to the I2C bus specification.
//----------------------------------------------------------------------------
byte i2c_read_data_size()
{
	// Output start condition
	if (!i2c_start())
		return 0;

	// Output I2C address (R/W bit low)
	if (!i2c_shift_out(i2c_address))
		return 0;

	// Output command CC_DATA_SIZE_READ
	if (!i2c_shift_out(CC_DATA_SIZE_READ))
		return 0;

	// As we have setup the read data size mode now, get the size itself
	return i2c_receive_byte();
}
//----------------------------------------------------------------------------
// Read info from the iLCD controller. i2c_result set to false in case of error.
// Data word info contains the status in the high byte and the # of bytes
// in output buffer in low byte.
// Done via a read word protocol due to the I2C bus specification.
//----------------------------------------------------------------------------
word i2c_read_info()
{
	// Output start condition
	if (!i2c_start())
		return 0;

	// Output I2C address (R/W bit low)
	if (!i2c_shift_out(i2c_address))
		return 0;

	// Output command CC_GET_INFO
	if (!i2c_shift_out(CC_GET_INFO))
		return 0;

	// As we have setup the read data size mode now, get the size itself
	return i2c_receive_word();
}
//----------------------------------------------------------------------------
// Read a block of data (max. 'count' bytes) from the iLCD controller.
// Returns # of bytes read. If any error occurs, -1 is returned. If the
// read buffer disposed by the caller ('count' length) is too small,
// count + 0x100 is returned.
//----------------------------------------------------------------------------
int i2c_read_block(byte *in_data, byte count)
{
	byte in_size;
	bool buff_too_small = false;

	// buffer must have a minimum size of 1 byte
	if (!count)
		return -1;

	// Output start condition
	if (!i2c_start())
		return -1;

	// Output I2C address (R/W bit low)
	if (!i2c_shift_out(i2c_address))
		return -1;

	// Output command CC_BLOCK_READ
	if (!i2c_shift_out(CC_BLOCK_READ))
		return -1;

	// Output start condition again to denote a read
	if (!i2c_start())
		return -1;

	// Output I2C address (R/W bit high)
	if (!i2c_shift_out(i2c_address | I2C_RW))
		return -1;

	// Get the data size available byte now
	in_size = i2c_shift_in();
	if (!i2c_result)
	{
		// Send Stop
		i2c_stop();

		// error shifting in data
		return -1;
	}

	if (in_size > count)
	{
		buff_too_small = true;
		in_size = count;
	}
	else
		count = in_size;

	for (; in_size; in_size--)
	{
		// ACK previous byte
		i2c_ack();

		// Get data byte and store to buffer
		*in_data++ = i2c_shift_in();

		if (!i2c_result)
		{
			// Send Stop
			i2c_stop();

			// error shifting in data
			return -1;
		}
	}

	// respond to last byte with NAK
	i2c_nak();

	// Send Stop
	i2c_stop();

	return count | (buff_too_small ? 0x100 : 0);
}
//----------------------------------------------------------------------------
// Write a block of data ('count' bytes) to the iLCD controller.
// Returns true if successful.
//----------------------------------------------------------------------------
bool i2c_write_block(byte *out_data, byte count)
{
	// count is not allowed to be zero
	if (!count)
		return false;

	// Output start condition
	if (!i2c_start())
		return false;

	// Output I2C address (R/W bit low)
	if (!i2c_shift_out(i2c_address))
		return false;

	// Output command CC_BLOCK_WRITE
	if (!i2c_shift_out(CC_BLOCK_WRITE))
		return false;

	// Output # of bytes to follow
	if (!i2c_shift_out(count))
		return false;

	while (count--)
	{
		if (!i2c_shift_out(*out_data++))
			return false;
	}

	i2c_stop();
	return true;
}
//----------------------------------------------------------------------------
// Receive byte from the iLCD controller. i2c_result set to false in case of error.
// Done via a receive byte protocol due to the I2C bus specification.
//----------------------------------------------------------------------------
static byte i2c_receive_byte()
{
	byte inbyte;

	// Output start condition
	if (!i2c_start())
		return 0;

	// Output I2C address (R/W bit high)
	if (!i2c_shift_out(i2c_address | I2C_RW))
		return 0;

	// Get the data byte now
	inbyte = i2c_shift_in();

	// Send NAK
	i2c_nak();

	// Send Stop
	i2c_stop();

	return inbyte;
}
//----------------------------------------------------------------------------
// Receive word from the iLCD controller. i2c_result set to false in case of error.
// Done via a receive word protocol due to the I2C bus specification.
//----------------------------------------------------------------------------
static word i2c_receive_word()
{
	data word wvalue;

	// Output start condition
	if (!i2c_start())
		return false;

	// Output I2C address (R/W bit high)
	if (!i2c_shift_out(i2c_address | I2C_RW))
		return false;

	// Get the first data byte now
	wvalue = i2c_shift_in() * 0x100;

	if (!i2c_result)
	{
		// Send Stop
		i2c_stop();

		return false;
	}

	// Send ACK
	i2c_ack();

	// Get the second data byte now
	wvalue |= i2c_shift_in();

	// If no more data is to be sent from iLCD controller,
	// a NAK has to sent from the master
	i2c_nak();

	// Send Stop
	i2c_stop();

	return wvalue;
}
//----------------------------------------------------------------------------
// Shift in a byte from the I2C device, most significant bit first.
// SCL_PIN expected low on entry. Return with SCL_PIN low.
// Byte shifted in is contained in static variable in_byte.
// Returns false in case of error.
//----------------------------------------------------------------------------
static byte i2c_shift_in()
{
	byte i, in_byte = 0;

	SDA_PIN = 1;	// make SDA_PIN an input
	delay();			// wait a while
	SCL_PIN = 1;	// raise clock
	in_byte = 0;

	// wait until slave releases clock
	for (i = 0xff; i; i--)
	{
		if (SCL_PIN)
			break;
	}

	if (!SCL_PIN)
	{
		i2c_result = false;
		return 0;
	}

	// shift in 8 bits - use i-- to allow using DJNZ instruction
	for (i = 8; i; i--)
	{
		SCL_PIN = 1;	// raise clock

		// read in bit into least significant byte position
		in_byte = (in_byte << 1) | SDA_PIN;

		SCL_PIN = 0;	// drop clock
	}

	i2c_result = true;
	return in_byte;
}
//----------------------------------------------------------------------------
// Shift out a byte to the I2C device, most significant bit first.
// SCL_PIN, SDA_PIN expected low on entry. Return with SCL_PIN low.
// Called with data to send.
// Returns false to indicate failure by slave to acknowledge or
// no releasing clock.
//----------------------------------------------------------------------------
static bool i2c_shift_out(byte data_byte)
{
	byte i, cnt;

	// shift out 8 bits - use i-- to allow using DJNZ instruction
	for (i = 8; i; i--)
	{
		SDA_PIN = data_byte & 0x80;	// output bit
		delay();								// wait a while
		SCL_PIN = 1;						// raise clock

		// wait until slave releases clock
		for (cnt = 0xff; cnt; cnt--)
		{
			if (SCL_PIN)
				break;
		}

		if (!SCL_PIN)
			return (i2c_result = false);	// clock not released

		SCL_PIN = 0;						// drop clock
		data_byte <<= 1;              // next bit
	}

	SDA_PIN = 1;							// release SDA_PIN for ACK
	delay();									// wait a while
	SCL_PIN = 1;							// raise ACK clock

	delay();									// wait a while
	i2c_result = (SDA_PIN == 0);		// get ACK bit
	SCL_PIN = 0;							// drop ACK clock

	return i2c_result;
}
//----------------------------------------------------------------------------
// Clock out an acknowledge bit (low).
// SCL_PIN expected low on entry. Return with SCL_PIN, SDA_PIN low.
// Note: There's no need to optimize this function via assembler,
// as Keil C works perfect here
//----------------------------------------------------------------------------
static void i2c_ack()
{
	SDA_PIN = 0;	// ACK bit
	delay();			// wait a while
	SCL_PIN = 1;	// raise clock
	delay();			// wait a while
	SCL_PIN = 0;	// drop clock
}
//----------------------------------------------------------------------------
// Clock out a negative acknowledge bit (high).
// SCL_PIN expected low on entry. Return with SCL_PIN low, SDA_PIN high.
// Note: There's no need to optimize this function via assembler,
// as Keil C works perfect here
//----------------------------------------------------------------------------
static void i2c_nak()
{
	SDA_PIN = 1;	// NAK bit
	delay();			// wait a while
	SCL_PIN = 1;	// raise clock
	delay();			// wait a while
	SCL_PIN = 0;	// drop clock
}
//----------------------------------------------------------------------------
// Send STOP, defined as low-to-high SDA_PIN with SCL_PIN high.
// SCL_PIN expected low on entry. Return with SCL_PIN, SDA_PIN high.
// Note: There's no need to optimize this function via assembler,
// as Keil C works perfect here
//----------------------------------------------------------------------------
static void i2c_stop()
{
	byte cnt;

	SDA_PIN = 0;
	delay();			// wait a while
	SCL_PIN = 1;

	// wait until slave releases clock
	for (cnt = 0xff; cnt; cnt--)
	{
		if (SCL_PIN)
			break;
	}

	SDA_PIN = 1;
}
//----------------------------------------------------------------------------
// Send START, defined as high-to-low SDA_PIN with SCL_PIN high.
// Return with SCL_PIN, SDA_PIN low.
// Returns false if bus is not available.
// Note: There's no need to optimize this function via assembler,
// as Keil C works perfect here
//----------------------------------------------------------------------------
static bool i2c_start()
{
	byte cnt;

	SDA_PIN = 1;
	delay();					// wait a while
	SCL_PIN = 1;

	// wait until bus is available
	for (cnt = 0xff; cnt; cnt--)
	{
		if (SDA_PIN && SCL_PIN)
		{
			SDA_PIN = 0;
			delay();			// wait a while
			SCL_PIN = 0;
			return true;
		}
	}

	return false;
}
//----------------------------------------------------------------------------
